"""Constants definitions related to tool wrappers."""

TOOL = "tool"
INPUTS = "inputs"
PARAM = "param"
CONDITIONAL = "conditional"
REPEAT = "repeat"
SECTION = "section"
NAME = "name"
TYPE = "type"
SELECT = "select"
OPTION = "option"
VALUE = "value"
WHEN = "when"
TEST = "test"
TESTS = "tests"
TEXT = "text"
MIN = "min"
MAX = "max"
INTEGER = "integer"
FLOAT = "float"
BOOLEAN = "boolean"
TRUEVALUE = "truevalue"
FALSEVALUE = "falsevalue"
CHECKED = "checked"
ARGUMENT = "argument"
BOOLEAN_OPTIONS = ["true", "false", "True", "False", "yes", "no", "1", "0"]
EXPECT_NUM_OUTPUTS = "expect_num_outputs"
OUTPUTS = "outputs"
DATA = "data"
COLLECTION = "collection"
OUTPUT = "output"
OUTPUT_COLLECTION = "output_collection"
ASSERT_CONTENTS = "assert_contents"
HAS_TEXT = "has_text"
HAS_LINE = "has_line"
HAS_LINE_MATCHING = "has_line_matching"
HAS_N_COLUMNS = "has_n_columns"
HAS_SIZE = "has_size"
LINE = "line"
EXPRESSION = "expression"
N = "n"
DELTA = "delta"
ELEMENT = "element"
DASH = "-"
UNDERSCORE = "_"
DESCRIPTION = "description"
XREF = "xref"
COMMAND = "command"
ENV_VARIABLES = "environment_variables"
CONFIGFILES = "configfiles"
OPTIONAL = "optional"
LABEL = "label"
HELP = "help"
FORMAT = "format"
MACROS = "macros"
IMPORT = "import"
TOKEN = "token"
MACRO = "macro"
XML = "xml"
